import { QueryBuilder } from "../QueryBuilder";
import { grpc } from "@improbable-eng/grpc-web";
import { Query } from "../generated/query_pb";
import { Response } from "../generated/response_pb";
import { QueryHeader } from "../generated/query_header_pb";
import { Hbar } from "../Hbar";
import { NftId } from "./NftId";
import { TokenId } from "./TokenId";
import { ResponseHeader } from "../generated/response_header_pb";
import { BaseClient } from "../BaseClient";
import { TokenNftInfo as ProtoTokenNftInfo, TokenGetNftInfoQuery } from "../generated/token_get_nft_info_pb";
import { TokenGetNftInfosQuery } from "../generated/token_get_nft_infos_pb";
import { TokenGetAccountNftInfosQuery } from "../generated/token_get_account_nft_infos_pb";
import { TokenService } from "../generated/token_service_pb_service";
import { AccountId } from "../account/AccountId";
import { timestampToDate } from "../Timestamp";

/**
 * Response when the client sends the node TokenGetNftInfoQuery.
 */
export interface TokenNftInfo {
    /**
     * The ID of the NFT
     */
    nftId: NftId | null;

    /**
     * The current owner of the NFT
     */
    accountId: AccountId | null;

    /**
     * The effective consensus timestamp at which the NFT was minted
     */
    creationTime: Date | null;

    /**
     * Represents the unique metadata of the NFT
     */
    metadata: Uint8Array | null;
}

/**
 * Get all the information about an token, including the balance.
 * This does not get the list of token records.
 */
export class TokenNftInfoQuery extends QueryBuilder<TokenNftInfo[]> {
    private readonly _byNftBuilder: TokenGetNftInfoQuery;
    private readonly _byTokenBuilder: TokenGetNftInfosQuery;
    private readonly _byAccountBuilder: TokenGetAccountNftInfosQuery;

    public constructor() {
        super();

        this._byNftBuilder = new TokenGetNftInfoQuery();
        this._byTokenBuilder = new TokenGetNftInfosQuery();
        this._byAccountBuilder = new TokenGetAccountNftInfosQuery();

        this._byNftBuilder.setHeader(new QueryHeader());
        this._byTokenBuilder.setHeader(new QueryHeader());
        this._byAccountBuilder.setHeader(new QueryHeader());
    }

    public setNftId(nftId: NftId): this {
        this._byNftBuilder.setNftid(nftId._toProto());

        this._inner.clearTokengetnftinfo();
        this._inner.clearTokengetnftinfos();
        this._inner.clearTokengetaccountnftinfos();

        this._inner.setTokengetnftinfo(this._byNftBuilder);
        return this;
    }

    public setTokenId(tokenId: TokenId): this {
        this._byTokenBuilder.setTokenid(tokenId._toProto());

        this._inner.clearTokengetnftinfo();
        this._inner.clearTokengetnftinfos();
        this._inner.clearTokengetaccountnftinfos();

        this._inner.setTokengetnftinfos(this._byTokenBuilder);
        return this;
    }

    public setAccountId(accountId: AccountId): this {
        this._byAccountBuilder.setAccountid(accountId._toProto());

        this._inner.clearTokengetnftinfo();
        this._inner.clearTokengetnftinfos();
        this._inner.clearTokengetaccountnftinfos();

        this._inner.setTokengetaccountnftinfos(this._byAccountBuilder);
        return this;
    }

    /**
     * Wrapper around `QueryBuilder.getCost()`. This must exist because the cost returned
     * `QueryBuilder.getCost()` and therein the Hedera Network doesn't work for any
     * acocuntns that have been deleted. In that case we want the minimum
     * cost to be ~25 Tinybar as this seems to succeed most of the time.
     */
    public async getCost(client: BaseClient): Promise<Hbar> {
        // deleted tokens return a COST_ANSWER of zero which triggers `INSUFFICIENT_TX_FEE`
        // if you set that as the query payment; 25 tinybar seems to be enough to get
        // `TOKEN_DELETED` back instead.
        const min = Hbar.fromTinybar(25);
        const cost = await super.getCost(client);
        return cost.isGreaterThan(min) ? cost : min;
    }

    protected _doLocalValidate(_: string[]): void {}

    protected _getMethod(): grpc.UnaryMethodDefinition<Query, Response> {
        return TokenService.getTokenNftInfo;
    }

    protected _getHeader(): QueryHeader {
        if (this._byAccountBuilder.hasAccountid()) {
            return this._byAccountBuilder.getHeader()!;
        } else if (this._byTokenBuilder.hasTokenid()) {
            return this._byTokenBuilder.getHeader()!;
        }
        return this._byNftBuilder.getHeader()!;
    }

    protected _mapResponseHeader(response: Response): ResponseHeader {
        if (this._byAccountBuilder.hasAccountid()) {
            return response.getTokengetaccountnftinfos()!.getHeader()!;
        } else if (this._byTokenBuilder.hasTokenid()) {
            return response.getTokengetnftinfos()!.getHeader()!;
        }
        return response.getTokengetnftinfo()!.getHeader()!;
    }

    protected _mapResponse(response: Response): TokenNftInfo[] {
        let infos: ProtoTokenNftInfo[];
        if (this._byAccountBuilder.hasAccountid()) {
            infos = response.getTokengetaccountnftinfos()!.getNftsList()!;
        } else if (this._byTokenBuilder.hasTokenid()) {
            infos = response.getTokengetnftinfos()!.getNftsList()!;
        } else {
            infos = [ response.getTokengetnftinfo()!.getNft()! ];
        }

        return infos.map((info) => ({
            nftId: info.hasNftid() ? NftId._fromProto(info.getNftid()!) : null,
            accountId: info.hasAccountid() ? AccountId._fromProto(info.getAccountid()!) : null,
            creationTime: info.hasCreationtime() ?
                timestampToDate(info.getCreationtime()!) :
                null,
            metadata: info.getMetadata_asU8()
        }));
    }
}
